function [theta, p, alpha, CF, E, elapsed] = MIQ(fp, f0, v0, degree)
    %function [theta, p, s, CF, E, elapsed] = nrosy_mex(fp, f0, v0, degree)
    %
    % Wraps the libigl MIQ implementation for nrosy field design.
    %
    % Input:
    %   fp - path to .off mesh file
    %   f0 - nC x 1 vector of constraint fids, where nC is the number of
    %        constraints.
    %   v0 - nC x 3 matrix of constraint vectors
    %	N  - field degree
    %
    % Output:
    %   theta   - nf x 1 vector with an angle per face
    %   p       - nFx1 vector of periods
    %   s       - nFx2 matrix of singularities (fid, s)
    %   CF      - nFx3 matrix of representative field vectors
    %   E       - MIQ energy
    %   elapsed - elapsed time in seconds
    %
    % Example:
    %   F0 = 1; V0 = [1, 0, 0]; DEGREE = 4;
    %   fp = './data/bunny.off';
    %   [theta, p, alpha, CF, E, elapsed] = MIQ(fp, F0, V0, DEGREE)
    %   figure
    %   m.plot(CF, alpha)

    if nargin < 4
        degree = 4;
    end

    cmd = ['[theta, p, alpha, CF, local_frames, frame_diffs, elapsed, pFixed] = ' ...
        'NRoSy_mex_bin(fp, f0, v0, degree);'];
    str = evalc(cmd);
    disp(str);
    
    m = Mesh(fp);
    [E, ~] = E_MIQ(m, theta, frame_diffs, p, degree);
    alpha = degree * alpha;
  
end

